import abc
from typing import Generic, TypeVar

R = TypeVar("R")


class BaseResource(abc.ABC, Generic[R]):
    """
    Decalarative resource definition.

    """

    def __init__(self, name: str, namespace: str):
        super().__init__()
        if "." in name:
            raise ValueError(
                "Resource cannot contain a dot(.) in their name", name
            )
        self.name = name
        self.namespace = namespace

    @property
    def canonical_name(self) -> str:
        """
        Canonical name of resource.

        It's <namespace>.<name>

        """
        return self.namespace + "." + self.name

    def __eq__(self, other):
        return (
            isinstance(other, BaseResource)
            and self.canonical_name == other.canonical_name
        )

    def __hash__(self):
        return hash((BaseResource, self.canonical_name))

    def __repr__(self):  # pragma: no cover
        return f"Resource({self.name!r}, {self.namespace!r})"



from typing import Any, Callable, ContextManager, Optional, TypeVar

from autowire.base_context import BaseContext
from autowire.base_resource import BaseResource
from autowire.builtins import context
from autowire.implementation import (
    ContextManagerImplementation,
    PlainFunctionImplementation,
)
from autowire.resource import Resource

R = TypeVar("R")


class Context(BaseContext):
    """
    Resource management context

    """

    def __init__(self, parent: Optional[BaseContext] = None):
        super().__init__(parent)
        self._set_builtins()

    def _set_builtins(self):
        # Current context
        @self.plain(context)
        def get_context():
            return self

    def plain(
        self,
        resource: Resource[R],
        *arg_resources: BaseResource[Any],
        **kwarg_resources: BaseResource[Any],
    ):
        """
        Provide resource's implementation with plain function

        arg_resources and kwarg_resources will be used for dependency injection.

        ::

            config = Resource("config", __name__)
            connection_pool = Resource("connection_pool", __name__)
            db_connection = Resource("db_connection", __name__)

            context = Context()

            @context.plain(db_connection, connection_pool, config=config)
            def get_db_connection(connection_pool: Pool, *, config: dict) -> Connection:
                # ...

        """

        def decorator(fn: Callable[..., R]):
            self.provide(
                resource,
                PlainFunctionImplementation(
                    fn, arg_resources, kwarg_resources
                ),
            )
            return fn

        return decorator

    def contextual(
        self,
        resource: Resource[R],
        *arg_resources: BaseResource[Any],
        **kwarg_resources: BaseResource[Any],
    ):
        """
        Provide resource's implementation with context manager

        arg_resources and kwarg_resources will be used for dependency injection.

        ::

            db_connection = Resource("db_connection", __name__)
            db_transaction = Resource("transaction", __name__)

            context = Context()

            @context.contextual(db_transaction, db_connection)
            @contextlib.contextmanager
            def begin_trasaction(db_connection: Connection):
                tx = db_connection.begin()
                try:
                    yield tx
                except Exception:
                    tx.rollback()
                    raise
                finally:
                    tx.commit()

        """

        def decorator(manager: Callable[..., ContextManager[R]]):
            self.provide(
                resource,
                ContextManagerImplementation(
                    manager, arg_resources, kwarg_resources
                ),
            )
            return manager

        return decorator



"""
autowire.builtins
=================

Builtin resources.

"""
from autowire.base_context import BaseContext
from autowire.resource import Resource

#: Resource that indicating current context
context: Resource[BaseContext] = Resource("context", __name__)



from __future__ import annotations

import abc
import contextlib
from typing import Any, Callable, ContextManager, Dict, Generic, Tuple, TypeVar

from autowire.base_resource import BaseResource
from autowire.provider import ResourceProvider

R = TypeVar("R")


class Implementation(Generic[R], metaclass=abc.ABCMeta):
    """
    Declarative base type for implementation

    """

    @abc.abstractmethod
    def reify(
        self,
        resource: BaseResource[R],
        provider: ResourceProvider,
    ) -> ContextManager[R]:  # pragma: no cover
        """
        Reify the resource

        """
        pass


class ContextManagerImplementation(Implementation[R]):
    """
    Use context manager as an implementaion

    """

    def __init__(
        self,
        manager_generator: Callable[..., ContextManager[R]],
        arg_resources: Tuple[BaseResource[Any], ...],
        kwarg_resources: Dict[str, BaseResource[Any]],
    ):
        super().__init__()
        self.manager_generator = manager_generator
        self.arg_resources = arg_resources
        self.kwarg_resources = kwarg_resources

    def reify(
        self,
        resource: BaseResource[R],
        provider: ResourceProvider,
    ):
        args = [provider.resolve(arg) for arg in self.arg_resources]
        kwargs = {
            name: provider.resolve(arg)
            for name, arg in self.kwarg_resources.items()
        }

        return self.manager_generator(*args, **kwargs)


class PlainFunctionImplementation(ContextManagerImplementation[R]):
    """
    Use plain function as an implementaion

    """

    def __init__(
        self,
        fn: Callable[..., R],
        arg_resources: Tuple[BaseResource[Any], ...],
        kwarg_resources: Dict[str, BaseResource[Any]],
    ):
        @contextlib.contextmanager
        def manager(*args, **kwargs):
            yield fn(*args, **kwargs)

        super().__init__(manager, arg_resources, kwarg_resources)
        self.fn = fn



import abc
from typing import TypeVar

from autowire.base_resource import BaseResource

R = TypeVar("R")


class ResourceProvider(abc.ABC):
    """
    Declarative base class for resource provider
    """

    @abc.abstractmethod
    def resolve(self, resource: BaseResource[R]) -> R:  # pragma: no cover
        pass



"""
autowire.base
=============

Base definitions of autowire.

"""
from __future__ import annotations

import collections
import sys
from typing import Any, ContextManager, Dict, Optional, Tuple, TypeVar, cast

from autowire.base_resource import BaseResource
from autowire.exc import ResourceNotProvidedError
from autowire.implementation import Implementation
from autowire.provider import ResourceProvider
from autowire.resource import Resource

R = TypeVar("R")


class BaseContext(ResourceProvider, ContextManager["BaseContext"]):
    """
    Resource management context base class

    """

    def __init__(self, parent: Optional[BaseContext] = None):
        super().__init__()
        self.parent = parent
        self.resource_pool: collections.OrderedDict[
            BaseResource[Any], Tuple[Any, ContextManager[Any]]
        ] = collections.OrderedDict()
        self.implementations: Dict[BaseResource[Any], Implementation[Any]] = {}

    def drain(self):
        """
        Drain all resources resolved by this context.

        """
        items = list(self.resource_pool.items())

        def drainer(
            items: list[BaseResource[Any], Tuple[Any, ContextManager[Any]]]
        ):
            if not items:
                return
            resource, (resolve, manager) = items.pop(0)
            try:
                drainer(items)
            except Exception:
                type_, value, traceback = sys.exc_info()
                self.resource_pool.pop(resource)
                manager.__exit__(type_, value, traceback)
                raise
            else:
                self.resource_pool.pop(resource)
                manager.__exit__(None, None, None)

        drainer(items)

    def provide(
        self, resource: Resource[R], implementation: Implementation[R]
    ):
        """
        Provide an implementation for resource.

        """
        self.implementations[resource] = implementation

    #
    # Resource provider implementation
    #

    def resolve(self, resource: BaseResource[R]) -> R:
        """
        Resolve resource in this context.

        """
        # Find resolved resource from resource pool
        if resource in self.resource_pool:
            resolved, manager = self.resource_pool[resource]
            return resolved

        # Find an implementation
        impl = None
        if resource in self.implementations:
            impl = self.implementations[resource]
        else:
            if self.parent is not None:
                # find from parent
                return self.parent.resolve(resource)
            elif (
                isinstance(resource, Resource)
                and resource.default_implementation is not None
            ):
                # use default implementation
                impl = resource.default_implementation
        if impl is None:
            # finally we fail
            raise ResourceNotProvidedError(
                "Resource not provided to this context",
                resource.canonical_name,
            )
        # resolve
        impl = cast(Implementation[R], impl)
        manager = impl.reify(resource, self)
        resolved = manager.__enter__()
        # throw into resource pool
        self.resource_pool[resource] = (resolved, manager)
        return resolved

    #
    # Context manager implementation
    #

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.drain()



from __future__ import annotations

from typing import Any, Callable, ContextManager, Optional, TypeVar

from autowire.base_resource import BaseResource
from autowire.implementation import (
    ContextManagerImplementation,
    Implementation,
    PlainFunctionImplementation,
)

R = TypeVar("R")


class Resource(BaseResource[R]):
    """
    Decalarative resource definition.

    Namespace generally be the module's name.

    `name` cannot include any dot(.) characters. ::

        >>> resource = BaseResource('name', __name__)

    """

    def __init__(self, name: str, namespace: str):
        super().__init__(name, namespace)
        self.default_implementation: Optional[Implementation] = None

    def plain(
        self,
        *arg_resources: BaseResource[Any],
        **kwarg_resources: BaseResource[Any],
    ):
        """
        Set default implementation with plain function

        arg_resources and kwarg_resources will be used for dependency injection.

        ::

            config = Resource("config", __name__)
            connection_pool = Resource("connection_pool", __name__)
            db_connection = Resource("db_connection", __name__)

            @db_connection.plain(connection_pool, config=config)
            def get_db_connection(connection_pool: Pool, *, config: dict) -> Connection:
                # ...

        """

        def decorator(fn: Callable[..., R]):
            self.default_implementation = PlainFunctionImplementation(
                fn, arg_resources, kwarg_resources
            )
            return fn

        return decorator

    def contextual(
        self,
        *arg_resources: BaseResource[Any],
        **kwarg_resources: BaseResource[Any],
    ):
        """
        Set default implementation with context manager

        arg_resources and kwarg_resources will be used for dependency injection.

        ::

            db_connection = Resource("db_connection", __name__)
            db_transaction = Resource("transaction", __name__)

            @db_transaction.contextual(db_connection)
            @contextlib.contextmanager
            def begin_trasaction(db_connection: Connection):
                tx = db_connection.begin()
                try:
                    yield tx
                except Exception:
                    tx.rollback()
                    raise
                finally:
                    tx.commit()

        """

        def decorator(manager: Callable[..., ContextManager[R]]):
            self.default_implementation = ContextManagerImplementation(
                manager, arg_resources, kwarg_resources
            )
            return manager

        return decorator



import setuptools
from setuptools import setup

setup(
    name="Autowire",
    packages=setuptools.find_packages(exclude=["tests"]),
    version="1.0.0",
    description="Simple dependency injection.",
    author="Choi Geonu",
    author_email="6566gun@gmail.com",
    url="https://github.com/hardtack/autowire",
    license="MIT LICENSE",
    keywords=["dependency-injection"],
)
